package com.ningdatech.pmapi.projectdeclared.manage;

import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.ningdatech.basic.function.VUtils;
import com.ningdatech.pmapi.common.enumeration.ProjectProcessStageEnum;
import com.ningdatech.pmapi.projectlib.enumeration.ProjectStatusEnum;
import com.ningdatech.pmapi.projectlib.model.entity.Project;
import com.ningdatech.pmapi.projectlib.service.IProjectInstService;
import com.ningdatech.pmapi.projectlib.service.IProjectService;
import com.wflow.bean.entity.WflowModels;
import com.wflow.exception.BusinessException;
import com.wflow.workflow.service.ProcessInstanceService;
import com.wflow.workflow.service.ProcessModelService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import java.util.Objects;

/**
 * @Classname ReviewByDeptJointManage
 * @Description 部门联审
 * @Date 2023/2/17 14:48
 * @Author PoffyZhang
 */
@Component
@Slf4j
@RequiredArgsConstructor
public class ReviewByDeptJointManage {

    private final IProjectService projectService;

    private final ProcessModelService processModelService;

    private final ProcessInstanceService processService;

    private final IProjectInstService projectInstService;

    /**
     * 省级部门联审
     * @param project
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Boolean startTheProcess(Project project) {
            VUtils.isTrue(Objects.isNull(project.getId())).throwMessage("提交失败 缺少项目ID!");
            Project projectInfo = projectService.getById(project.getId());
            VUtils.isTrue(Objects.isNull(projectInfo)).throwMessage("提交失败 此项目不存在!");
            String regionCode = projectInfo.getAreaCode();

            WflowModels model = processModelService.getOne(Wrappers.lambdaQuery(WflowModels.class)
                    .eq(WflowModels::getRegionCode, regionCode)
                    .eq(WflowModels::getProcessType, ProjectProcessStageEnum.DEPARTMENT_JOINT_APPROVAL_PROCESS.getCode())
                    .last("limit 1"));

            if (Objects.isNull(model)) {
                log.error("此 【{}】区域找不到 部门联审申报流程配置", regionCode);
                throw new BusinessException(String.format("此 【%s】区域找不到 部门联审申报流程配置", regionCode));
            }
            //要判断 项目当前状态 是不是 部门联审
            VUtils.isTrue(!ProjectStatusEnum.DEPARTMENT_JOINT_REVIEW.getCode().equals(projectInfo.getStatus()) ||
                            !ProjectStatusEnum.NOT_APPROVED.getCode().equals(projectInfo.getStage()))
                    .throwMessage("提交失败 该项目不是 省级部门联审状态状态或者未立项阶段");

                // 再判断 该项目是否 真实走完 预审审批 并且提取出 提交人
//            projectInstService.
//
//            ProcessStartParamsVo params = new ProcessStartParamsVo();
//            params.setUser(dto.getUser());
//            params.setProcessUsers(Collections.emptyMap());
//            //放入条件判断的项目字段
//            ProjectConditionDTO conditionDto = new ProjectConditionDTO();
//            BeanUtils.copyProperties(projectInfo, conditionDto);
//            dto.getFormData().putAll(
//                    JSON.parseObject(JSON.toJSONString(conditionDto), new TypeReference<Map<String, Object>>() {
//                    })
//            );
//            params.setFormData(dto.getFormData());
//            String instanceId = processService.startProcess(model.getProcessDefId(), params);
//            log.info("建设方案项目申报成功 【{}】", instanceId);
//
//            //保存建设项目
//            modifyProject(projectInfo, instanceId, projectInfo.getConstructionPlanFile());

            return null;
    }
}
